@@ -29,11 +29,11 @@
 ****************************************************************************/
 
 #include <QtQuick3DRuntimeRender/private/qssgrenderloadedtexture_p.h>
-#include <QtQuick3DRuntimeRender/private/qssgrenderinputstreamfactory_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendererutil_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgruntimerenderlogging_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendertexturedata_p.h>
 #include <QtGui/QImageReader>
+#include <QtGui/QColorSpace>
 #include <QtMath>
 
 #include <QtQuick3DUtils/private/qssgutils_p.h>
@@ -42,6 +42,76 @@
 
 QT_BEGIN_NAMESPACE
 
+
+QSharedPointer<QIODevice> QSSGInputUtil::getStreamForFile(const QString &inPath, bool inQuiet, QString *outPath)
+{
+    QFile *file = nullptr;
+    QString tryPath = inPath.startsWith(QLatin1String("qrc:/")) ? inPath.mid(3) : inPath;
+    QFileInfo fi(tryPath);
+    bool found = fi.exists();
+    if (!found && fi.isNativePath()) {
+        tryPath.prepend(QLatin1String(":/"));
+        fi.setFile(tryPath);
+        found = fi.exists();
+    }
+    if (found) {
+        QString filePath = fi.canonicalFilePath();
+        file = new QFile(filePath);
+        if (file->open(QIODevice::ReadOnly)) {
+            if (outPath)
+                *outPath = filePath;
+        } else {
+            delete file;
+            file = nullptr;
+        }
+    }
+    if (!file && !inQuiet)
+        qCWarning(WARNING, "Failed to find file: %s", qPrintable(inPath));
+    return QSharedPointer<QIODevice>(file);
+}
+
+QSharedPointer<QIODevice> QSSGInputUtil::getStreamForTextureFile(const QString &inPath, bool inQuiet,
+                                                                 QString *outPath, FileType *outFileType)
+{
+    static const QList<QByteArray> hdrFormats = QList<QByteArray>({ "hdr" });
+    static const QList<QByteArray> textureFormats = QTextureFileReader::supportedFileFormats();
+    static const QList<QByteArray> imageFormats = QImageReader::supportedImageFormats();
+    static const QList<QByteArray> allFormats = textureFormats + hdrFormats + imageFormats;
+
+    QString filePath;
+    QByteArray ext;
+    QSharedPointer<QIODevice> stream = getStreamForFile(inPath, true, &filePath);
+    if (stream) {
+        ext = QFileInfo(filePath).suffix().toLatin1().toLower();
+    } else {
+        for (const QByteArray &format : allFormats) {
+            QString tryName = inPath + QLatin1Char('.') + QLatin1String(format);
+            stream = getStreamForFile(tryName, true, &filePath);
+            if (stream) {
+                ext = format;
+                break;
+            }
+        }
+    }
+    if (stream) {
+        if (outPath)
+            *outPath = filePath;
+        if (outFileType) {
+            FileType type = UnknownFile;
+            if (hdrFormats.contains(ext))
+                type = HdrFile;
+            else if (textureFormats.contains(ext))
+                type = TextureFile;
+            else if (imageFormats.contains(ext))
+                type = ImageFile;
+            *outFileType = type;
+        }
+    } else if (!inQuiet) {
+        qCWarning(WARNING, "Failed to find texture file for: %s", qPrintable(inPath));
+    }
+    return stream;
+}
+
 static inline QSSGRenderTextureFormat fromGLtoTextureFormat(quint32 internalFormat)
 {
     switch (internalFormat) {
@@ -218,9 +288,9 @@ static inline QSSGRenderTextureFormat fromGLtoTextureFormat(quint32 internalForm
 
 QSSGLoadedTexture *QSSGLoadedTexture::loadQImage(const QString &inPath, qint32 flipVertical)
 {
-    Q_UNUSED(flipVertical);
-
     QImage image(inPath);
+    if (image.isNull())
+        return nullptr;
     const QPixelFormat pixFormat = image.pixelFormat();
     QImage::Format targetFormat = QImage::Format_RGBA8888_Premultiplied;
     if (image.colorCount()) // a palleted image
@@ -233,7 +303,8 @@ QSSGLoadedTexture *QSSGLoadedTexture::loadQImage(const QString &inPath, qint32 f
         targetFormat = QImage::Format_RGBA8888;
 
     image.convertTo(targetFormat); // convert to a format mappable to QRhiTexture::Format
-    image.mirror(); // Flip vertically to the conventional Y-up orientation
+    if (flipVertical)
+        image.mirror(); // Flip vertically to the conventional Y-up orientation
 
     QSSGLoadedTexture *retval = new QSSGLoadedTexture;
     retval->width = image.width();
@@ -243,14 +314,14 @@ QSSGLoadedTexture *QSSGLoadedTexture::loadQImage(const QString &inPath, qint32 f
     retval->data = (void *)retval->image.bits();
     retval->dataSizeInBytes = image.sizeInBytes();
     retval->setFormatFromComponents();
+    // #TODO: This is a very crude way detect color space
+    retval->isSRGB = image.colorSpace().transferFunction() != QColorSpace::TransferFunction::Linear;
 
     return retval;
 }
 
-QSSGLoadedTexture *QSSGLoadedTexture::loadCompressedImage(const QString &inPath, bool inFlipY)
+QSSGLoadedTexture *QSSGLoadedTexture::loadCompressedImage(const QString &inPath)
 {
-    Q_UNUSED(inFlipY);
-
     QSSGLoadedTexture *retval = nullptr;
 
     // Open File
@@ -267,14 +338,14 @@ QSSGLoadedTexture *QSSGLoadedTexture::loadCompressedImage(const QString &inPath,
         return retval;
     }
     retval = new QSSGLoadedTexture;
-    retval->compressedData = reader->read();
+    retval->textureFileData = reader->read();
 
     // Fill out what makes sense, leave the rest at the default 0 and null.
-    retval->width = retval->compressedData.size().width();
-    retval->height = retval->compressedData.size().height();
-    auto glFormat = retval->compressedData.glInternalFormat()
-            ? retval->compressedData.glInternalFormat()
-            : retval->compressedData.glFormat();
+    retval->width = retval->textureFileData.size().width();
+    retval->height = retval->textureFileData.size().height();
+    auto glFormat = retval->textureFileData.glInternalFormat()
+            ? retval->textureFileData.glInternalFormat()
+            : retval->textureFileData.glFormat();
     retval->format = fromGLtoTextureFormat(glFormat);
 
     delete reader;
@@ -364,7 +435,7 @@ void decodeScanlineToTexture(RGBE *scanline, int width, void *outBuf, quint32 of
     target += offset;
 
     if (inFormat == QSSGRenderTextureFormat::RGBE8) {
-        memcpy(target, scanline, size_t(4 * width));
+        memcpy(target, scanline, size_t(width) * 4);
     } else {
         float rgbaF32[4];
         for (int i = 0; i < width; ++i) {
@@ -466,6 +537,7 @@ QSSGLoadedTexture *QSSGLoadedTexture::loadHdrImage(const QSharedPointer<QIODevic
     imageData->height = height;
     imageData->format = format;
     imageData->components = format.getNumberOfComponent();
+    imageData->isSRGB = false;
 
     // Allocate a scanline worth of RGBE data
     RGBE *scanline = new RGBE[width];
@@ -503,6 +575,18 @@ QSSGLoadedTexture *QSSGLoadedTexture::loadTextureData(QSSGRenderTextureData *tex
     imageData->format = textureData->format();
     imageData->components = textureData->format().getNumberOfComponent();
 
+    // #TODO: add an API to make this explicit
+    // For now we assume HDR formats are linear and everything else
+    // is sRGB, which is not ideal but so far this is only used by
+    // the environment mapper code
+    if (imageData->format == QSSGRenderTextureFormat::RGBE8 ||
+        imageData->format == QSSGRenderTextureFormat::RGBA16F ||
+        imageData->format == QSSGRenderTextureFormat::RGBA32F ||
+        imageData->format == QSSGRenderTextureFormat::BC6H)
+        imageData->isSRGB = false;
+    else
+        imageData->isSRGB = true;
+
     return imageData;
 }
 
@@ -606,28 +690,30 @@ bool QSSGLoadedTexture::scanForTransparency() const
 
 QSSGLoadedTexture *QSSGLoadedTexture::load(const QString &inPath,
                                            const QSSGRenderTextureFormat &inFormat,
-                                           QSSGInputStreamFactory &inFactory,
                                            bool inFlipY)
 {
     if (inPath.isEmpty())
         return nullptr;
 
     QSSGLoadedTexture *theLoadedImage = nullptr;
-    QSharedPointer<QIODevice> theStream(inFactory.getStreamForFile(inPath));
     QString fileName;
-    inFactory.getPathForFile(inPath, fileName);
-    if (theStream && inPath.size() > 3) {
-        QByteArray ext = inPath.mid(inPath.lastIndexOf(QLatin1Char('.')) + 1).toLower().toUtf8();
-        if (ext == "hdr") {
+    QSSGInputUtil::FileType fileType = QSSGInputUtil::UnknownFile;
+    QSharedPointer<QIODevice> theStream =
+            QSSGInputUtil::getStreamForTextureFile(inPath, true, &fileName, &fileType);
+
+    if (theStream) {
+        switch (fileType) {
+        case QSSGInputUtil::HdrFile:
             // inFormat is a suggestion that's only relevant for HDR images
             // (tells if we want want RGBA16F or RGBE-on-RGBA8)
             theLoadedImage = loadHdrImage(theStream, inFormat);
-        } else if (QTextureFileReader::supportedFileFormats().contains(ext)) {
-            theLoadedImage = loadCompressedImage(fileName, inFlipY);
-        } else if (QImageReader::supportedImageFormats().contains(ext)) {
+            break;
+        case QSSGInputUtil::TextureFile:
+            theLoadedImage = loadCompressedImage(fileName); // no choice but to ignore inFlipY here
+            break;
+        default:
             theLoadedImage = loadQImage(fileName, inFlipY);
-        } else {
-            qCWarning(INTERNAL_ERROR, "Unrecognized image extension: %s", qPrintable(inPath));
+            break;
         }
     }
     return theLoadedImage;
